import os
from enum import Enum
from typing import List

from on_rails import Result, ValidationError, def_result

from pylity.Collection import Collection
from pylity.String import String


class PathType(Enum):
    """
    Defines an enumeration of three path types: file, directory, and invalid.
    """
    FILE = 1
    DIRECTORY = 2
    INVALID = 3


class Path:
    """ A collection of utility functions for paths  """

    @staticmethod
    @def_result()
    def basename(path: str) -> Result[str]:
        """
        Takes a path as input and returns the name as a Result object, with an error message if the
        input is invalid.

        :param path: A string representing the path of file, directory, etc.
        :type path: str

        :return: Returns a `Result` object that contains either a string
        representing the name or a `ValidationError` object if the input path is not valid.
        """

        if String.is_none_or_empty(path):
            return Result.fail(ValidationError(
                message="The input file path is not valid. It can not be None or empty."))
        return Result.ok(os.path.basename(path))

    @staticmethod
    @def_result()
    def get_path_type(path: str) -> Result[PathType]:
        """
        The function checks the type of given path and returns a result indicating whether it is a
        file, directory, or invalid.

        :param path: A string representing a file path or directory path
        :type path: str

        :return: a `Result` object that contains a `PathType` value.
        """
        if not isinstance(path, str):
            return Result.fail(ValidationError(
                message=f"The input is not valid. Expected get a string but got {type(path)}"))

        if String.is_none_or_empty(path):
            return Result.ok(value=PathType.INVALID)

        if os.path.isfile(path):
            return Result.ok(value=PathType.FILE)
        if os.path.isdir(path):
            return Result.ok(value=PathType.DIRECTORY)
        return Result.ok(value=PathType.INVALID)

    @staticmethod
    @def_result()
    def collect_files(files_and_directories: List[str]) -> Result[List[str]]:
        """
        The function collects all files from a list of directories and files and returns a list of the file paths.

        :param files_and_directories: A list of strings representing file paths and/or directory paths
        :type files_and_directories: List[str]

        :return: The function `collect_files` returns a `Result` object that contains a list of strings representing the
        collected files and directories.
        """

        files_and_directories = files_and_directories or []
        if not Collection.is_list(files_and_directories, str):
            return Result.fail(ValidationError(message="Input is not valid. It must be a list of strings."))

        files = set()

        for file_or_dir in files_and_directories:
            result = Path.get_path_type(file_or_dir) \
                .on_fail_break_function() \
                .on_success_fail_when(lambda path: path is PathType.INVALID,
                                      ValidationError(title="File or directory is not valid.",
                                                      message=f"The ({file_or_dir}) is not valid.")) \
                .on_fail_break_function()
            if result.value is PathType.FILE:
                files.add(file_or_dir)
            else:
                Path.collect_files_from_dir(file_or_dir).on_fail_break_function() \
                    .on_success(lambda dir_files: files.update(dir_files)).on_fail_break_function()
        return Result.ok(value=list(files))

    @staticmethod
    @def_result()
    def collect_files_from_dir(directory: str) -> Result[List[str]]:
        """
        This function collects all files from a given directory and returns them as a list.

        :param directory: The directory parameter is a string that represents the path to a directory from
        which we want to collect files
        :type directory: str
        """

        return Path.get_path_type(directory) \
            .on_success_fail_when(lambda path: path is not PathType.DIRECTORY,
                                  ValidationError(title="Directory is not valid.",
                                                  message=f"The ({directory}) is not valid.")) \
            .on_success(lambda: Path.__collect_files_from_dir(directory))

    @staticmethod
    @def_result()
    def __collect_files_from_dir(directory: str) -> Result[List[str]]:
        files = []
        for filename in os.listdir(directory):
            full_path = os.path.join(directory, filename)
            if os.path.isfile(full_path):
                files.append(full_path)
        return Result.ok(files)



class String:
    """ A collection of utility functions for string  """

    @staticmethod
    def is_none_or_empty(string: str) -> bool:
        """
        The function checks if a given string is None or empty (contains only whitespace characters)

        :param string: A string variable that we want to check if it is None or empty
        :type string: str

        :return: Returns a boolean value.
        """
        return string is None or string.strip() == ''



import asyncio
from asyncio import AbstractEventLoop

from on_rails import Result, def_result


class Async:
    """ A collection of utility functions for asynchronously  """

    @staticmethod
    @def_result()
    def get_loop() -> Result[AbstractEventLoop]:
        """
        The function returns the current event loop or creates a new one if none exists.

        :return: Returns a `Result` object that contains either an `AbstractEventLoop` instance or
        an error message.
        """
        try:
            return Result.ok(value=asyncio.get_event_loop())
        except RuntimeError as e:  # pragma: no cover
            if str(e).startswith('There is no current event loop in thread'):
                loop = asyncio.new_event_loop()
                asyncio.set_event_loop(loop)
                return Result.ok(value=loop)
            raise  # pragma: no cover



from typing import Optional, Type


class Collection:
    """
    A collection of utility functions for Collections like list, set, dict, etc.
    """

    @staticmethod
    def is_list(target, item_types: Optional[Type] = None) -> bool:
        """
        This function checks if a given target is a list and optionally checks if all its items are of a specified type.

        :param target: The variable or object that is being checked to see if it is a list

        :param item_types: The parameter `item_types` is an optional argument that specifies the type of
        items that should be present in the list. If `item_types` is provided, the function checks if all
        the items in the list are of the specified type. If `item_types` is not provided, the function only
        checks if target is list or not.
        :type item_types: Optional[Type]
        """

        if item_types:
            return isinstance(target, list) and all(isinstance(item, str) for item in target)
        return isinstance(target, list)



import inspect
from typing import Any, Callable, Dict, Optional

from on_rails import Result, ValidationError, def_result, try_func
from schema import Schema, SchemaError


def validate_func_params(schema: Optional[Schema] = None, raise_exception: bool = False):
    """
    Decorator function that validates function parameters based on a given schema and can optionally raise
    exceptions or manage exceptions with on_rails package.

    :param schema: An optional parameter of type Schema that specifies the schema to be used for validating the function
    parameters. If not provided, no validation will be performed
    :type schema: Optional[Schema]

    :param raise_exception: A boolean parameter that determines whether or not to raise an exception if the validation
    fails. If set to True, an exception will be raised. If set to False, the validation result will be returned as
    Result object, defaults to False
    :type raise_exception: bool (optional)

    :return: By default, when raise_exception is false, returns result as Result object, Otherwise returns output
    of function or raise exception.
    """

    def decorator(func: Callable):
        def wrapper(*args, **kwargs):
            # Validation
            validation_result = _get_args(func, *args, **kwargs) \
                .on_success(lambda arg_dict: _validate_args(schema, arg_dict, func))

            if validation_result.success:
                if raise_exception:
                    # Execute the function without catch exceptions
                    return func(*args, **kwargs)
                # Execute the function with catch exceptions
                return try_func(lambda: func(*args, **kwargs))

            if not raise_exception:
                return validation_result

            if validation_result.detail and validation_result.detail.is_instance_of(ValidationError):
                raise ValueError(validation_result.detail.message)

            validation_result.on_fail_raise_exception()  # pragma: no cover
            raise Exception("Somethings is not correct. This code should not be executed!")  # pragma: no cover

        return wrapper

    return decorator


def _has_exception(result: Result):
    return result and result.detail and result.detail.exception


def _is_exception_type(result: Result, exception_type):
    return _has_exception(result) and isinstance(result.detail.exception, exception_type)


@def_result()
def _get_args(func: Callable, *args, **kwargs) -> Result[Dict[str, Any]]:
    """
    This function takes a function and its arguments and returns a dictionary of argument names and values, including any
    default values.

    :param func: The function for which we want to get the arguments and their values.
    :type func: Callable
    """

    # Get the function signature
    sig = inspect.signature(func)

    # Create a dictionary of argument names and values
    bound_args = try_func(lambda: sig.bind(*args, **kwargs)) \
        .on_fail_operate_when(lambda res: _is_exception_type(res, TypeError),
                              lambda res: Result.fail(ValidationError(message=str(res.detail.exception)))) \
        .on_fail_break_function() \
        .value

    arg_dict = {}
    for name, value in bound_args.arguments.items():
        arg_dict[name] = value

    # Add any additional keyword arguments to the argument dictionary
    arg_dict.update(kwargs)

    for param in sig.parameters.values():
        # If parameter is not in arg_dict and has default value, add it to the arg_dict
        if param.name not in arg_dict and param.default is not inspect.Parameter.empty:
            arg_dict[param.name] = param.default

    return Result.ok(value=arg_dict)


@def_result()
def _validate_args(schema: Optional[Schema], arg_dict: Dict[str, Any], func: Callable) -> Result:
    """
    This function validates arguments using a schema and returns a result, handling any validation errors.

    :param schema: An optional parameter of type Schema, which is used to validate the arguments passed to the function
    :type schema: Optional[Schema]

    :param arg_dict: arg_dict is a dictionary containing the arguments passed to a function. The keys of the dictionary are
    the argument names and the values are the argument values
    :type arg_dict: Dict[str, Any]

    :param func: Represents the function that needs to be validated. The function's annotations are
    used to determine the expected types of the arguments
    :type func: Callable
    """

    func_schema = schema if schema is not None else Schema(
        {key: value for key, value in arg_dict.items() if key in func.__annotations__})

    return try_func(lambda: func_schema.validate(arg_dict)) \
        .on_fail_operate_when(lambda res: _is_exception_type(res, SchemaError),
                              lambda res: Result.fail(ValidationError(message=str(res.detail.exception))))



import asyncio
import inspect
from typing import Callable, Coroutine

from on_rails import Result, ValidationError, def_result

from pylity.Async import Async


class Function:
    """ A collection of utility functions for functions  """

    @staticmethod
    def is_func_valid(func) -> bool:
        """
        Checks if a given input is a valid callable function or not.

        :param func: The parameter `func` is expected to be a function object. This function checks if the `func`
        parameter is not `None` and is callable (i.e., it can be called as a function).
        If both conditions are true, it returns `True`

        :return: Returns a boolean value. It returns `True` if the input `func` is not `None` and
        is callable (i.e. can be called as a function), and `False` otherwise.
        """
        return func is not None and callable(func)

    @staticmethod
    @def_result()
    def get_num_of_params(func: Callable) -> Result[int]:
        """
        The function takes a callable object and returns the number of parameters it accepts.
        The number of parameters of some builtin functions cannot be recognized and
        returns ErrorDetail for these functions.

        :param func: The input function for which we want to determine the number of parameters.
        It is expected to be a callable object
        :type func: Callable

        :return: Returns a `Result` object that contains either the number of parameters of the input function
        as an integer if the function is valid, or a `ValidationError` message if the input function is not valid.
        """

        if not Function.is_func_valid(func):
            return Result.fail(ValidationError(message="The input function is not valid."))

        try:
            return Result.ok(len(inspect.signature(func).parameters))
        except ValueError:
            try:
                return Result.ok(func.__code__.co_argcount)
            except Exception:
                return Result.fail(ValidationError(title="Parameter Number Detection Error",
                                                   message=f"Can not determine the number of parameters of the "
                                                           f"{func.__name__}() function. You can use python function"
                                                           f"like lambda to fix this issue."))

    @staticmethod
    @def_result()
    def is_async(func: Callable) -> Result[bool]:
        """
        The function checks if a given function is asynchronous or not.

        :param func: The parameter `func` is a function that is being checked for whether it is an asynchronous
        function or not. The function is expected to be a valid Python function that can be called.
        :type func: Callable

        :return: Returns a boolean value indicating whether the input function is an asynchronous coroutine
        function or not. If the input function is not valid, it returns a `Result` object with a validation error.
        message.
        """

        if not Function.is_func_valid(func):
            return Result.fail(ValidationError(message="The input function is not valid."))

        return Result.ok(value=asyncio.iscoroutinefunction(func))

    @staticmethod
    @def_result()
    def await_func(func: callable) -> Result:
        """
        The function takes in a callable function, checks if it's valid, and runs it either synchronously and
        returns result as Result object.

        :param func: The parameter `func` is a callable object, which means it can be a function, method, or any other
        object that can be called like a function
        :type func: callable

        :return: Returns an instance of the `Result` class. If the input function is not valid, it returns
        a failed `Result` object with a `ValidationError` message. If the input function returns a coroutine, it
        uses the `run_until_complete` method of the event loop to wait for the coroutine to complete and returns
        the result as a `Result` object.
        """

        if not Function.is_func_valid(func):
            return Result.fail(ValidationError(message="The input function is not valid."))

        result = func()
        if isinstance(result, Coroutine):
            return Async.get_loop() \
                .on_success(lambda loop: loop.run_until_complete(result)) \
                .on_fail_break_function()

        return Result.convert_to_result(result)
